/*****************************************************************
* Karakteres LCD meghajtó programkönyvtár.                       *
* Specialitása, hogy a begin() feltölti a kijelzőre a magyar     *
* ékezetes karaktereket (kisbetűs formában). A byte típusú       * 
* változók DEC,HEX,BIN formátumban is megjeleníthetők            *
* További részletek a www.STARduino.hu weboldalon a "Modulok /   *
* kijelzők / 1602 LCD kijelző" menüpontban                       *
*****************************************************************/

#include "Arduino.h"
#include "Wire.h"
#include "LiquidCrystal_I2C_HUN.h"
//#include <inttypes.h>


LiquidCrystal_I2C_HUN::LiquidCrystal_I2C_HUN(byte lcd_cim,byte lcd_oszlop,byte lcd_sor)
{
  _lcd_cim=lcd_cim;
  _lcd_oszlop=lcd_oszlop;
  _lcd_sor=lcd_sor;
  _backlightval = B00000000;   //háttérvilágítás alapértelmezetten kikapcsolva (I2C illesztő chip P3 port=0)
  _displaycontrol = B00001100; //utasítás regiszter (IR) alaphelyzet: display be, cursor ki, blink ki
                               //jobbról 1.bit blink be-ki, 2.bit cursor be-ki, 3.bit display be-ki.
                               //Jobbról a 4.bit 1, mert ezeknek a beállító parancsa ezzel a bittel kezdődik.
  _displaymode = B00000110;    // alapértelmezett cursor mozgatási irány (jobbra) és képernyő tartalom eltolás (csökkenő).
                               //Ezeknek nincs közvetlen hatása, de a következő tartalom kiírásakor e szerint fog változni
                               //a cursor pozíció, illetve a képernyő eltolás (ha be vannak kapcsolva).
                               // irány beállítása. Jobbról 3. a parancs jelzése, 2.bit cursor mozgatási irány (jobbra 1, balra 0), ű
                               // 1. bit képernyő tartalom eltolási irány (növekvő 1, csökkenő 0)
}

void LiquidCrystal_I2C_HUN::begin() {
  // Kijelző bekapcsolási procedura, alap konfiguráció:
  // 1. Átállítás 4-bit írási módba
  // 2. Működési funkció beállítások:
  //    DL=0; 8 bites interfész adatok
  //    N=0; 1 soros kijelző vagy N=1 két soros kijelző
  //    F=0; 5x8 pont karakterbetűtípus
  // 3. Kijelző control alaphelyzet:
  //    D = 1; Kijelző bekapcsolva
  //    C = 0; Kurzor kikapcsolva
  //    B = 0; A villogás kikapcsolva
  // 4. Kijelző törlés
  // 5. Belépési mód beállítása:
  //    I/D = 1; Írás után a DDRAM cím növekszik (jobbra ír)
  //    S = 0; Nincs képernyő eltolás
  // 6. Home funkcióba kerül a cursor (legkiesebb DDRAM cím)
  // 7. Háttérvilágítás kikapcsolva
  // 8. Magyar ékezetes karakterképek letöltése a CGRAM-ba
  delay(50);  //az adatlap szerint legalább 40 ms-ra van szükségünk a parancsok küldése előtt, miután a tápfesz 2,7 V fölé emelkedik
  // 8 bites parancsmódban indul a chip, ezért első lépésben át kell kapcsolni 4 bites módba. 
  // Az átkapcsoláshoz 3x kell kiküldeni. Az átkapcsoláshoz 2x kell, de megszakadhat
  // a folyamat előzőleg, ezért 3x. Ez az adatlap ajánlása, bár elvileg kétszer is elég lenne.
  wr4bit(0x03 << 4); //"0011" feltolva a D7-D4 bitekre (Functio set parancs: B00110000, DL=1 8 bites mód) 
  delayMicroseconds(4500); // Minimális végrehajtási idő 4.1ms
  wr4bit(0x03 << 4);
  delayMicroseconds(4500); // Minimális végrehajtási idő 4.1ms
  wr4bit(0x03 << 4);
  delayMicroseconds(150);
  wr4bit(0x02 << 4);                        //DL=0 4 bites mód (ez állítja be a 4 bites módot, function set parancs: B00100000)
  if (_lcd_sor > 1) {command(B00101000);}   //több soros kijelző esetén alapértelmezetten  4 bites mód , 2 soros mód, 5x8 karaktermap mód
  else {command(B00100000);}                //egy soros kijelző esetén alapértelmezetten  4 bites mód , 1 soros mód, 5x8 karaktermap mód
  display();                                //default érték beállítása (kijelző on, cursor off, blink off) a _displaycontrol 
                                            //változót írja ki a chip-re parancs módban (RS kivezetés=0)
  clear();                                  //Képernyő törlés (RS kivezetés=0, kiküldött byte B00000001
  command(_displaymode);                    //corsor mozgatás jobbra, képernyő eltolás csökkenő beállítása
  home();                                   //cursor a 0,0 pozícióba
  noBacklight();                            //háttérvilágítás ki
  //Hiányzó magyar ékezetes karakterek feltöltése a CGRAM-ba
  byte tt[7] = {B10,B100,B1110,B1,B1111,B10001,B1111};                                   //á betű karakterképe
  createChar(0,tt);                                                                      //betöltés a 0 című CGRAM byte-ba
  tt[0]=B10;tt[1]=B100;tt[2]=B1110;tt[3]=B10001;tt[4]=B11111;tt[5]=B10000;tt[6]=B1110;   //é betű karakterképe
  createChar(1,tt);                                                                      //betöltés a 1 című CGRAM byte-ba
  tt[0]=B10;tt[1]=B100;tt[2]=B0;tt[3]=B1110;tt[4]=B100;tt[5]=B100;tt[6]=B1110;           //í betű karakterképe
  createChar(2,tt);                                                                      //betöltés a 2 című CGRAM byte-ba
  tt[0]=B10;tt[1]=B100;tt[2]=B0;tt[3]=B1110;tt[4]=B10001;tt[5]=B10001;tt[6]=B1110;       //ó betű karakterképe
  createChar(3,tt);                                                                      //betöltés a 3 című CGRAM byte-ba
  tt[0]=B00101;tt[1]=B01010;tt[2]=B0;tt[3]=B1110;tt[4]=B10001;tt[5]=B10001;tt[6]=B1110;  //ő betű karakterképe
  createChar(4,tt);                                                                      //betöltés a 4 című CGRAM byte-ba
  tt[0]=B10;tt[1]=B100;tt[2]=B10001;tt[3]=B10001;tt[4]=B10001;tt[5]=B10011;tt[6]=B1101;  //ú betű karakterképe
  createChar(5,tt);                                                                      //betöltés a 5 című CGRAM byte-ba
  tt[0]=B00101;tt[1]=B01010;tt[2]=B0;tt[3]=B10001;tt[4]=B10001;tt[5]=B10011;tt[6]=B1101; //ű betű karakterképe
  createChar(6,tt);                                                                      //betöltés a 6 című CGRAM byte-ba
}

void LiquidCrystal_I2C_HUN::clear(){
  command(B00000001);       //Kiküldjük a chip-re a tartalmat, miközben RS kivezetés = 0 
  delayMicroseconds(2000);  //Végrehajtási idő adatlap szerint 1.5msec. Cursor pozíciót is beállítja 0-ra.
}

void LiquidCrystal_I2C_HUN::home(){
  command(B00000010);       //A cursor pozíciót 0-ra (0 sor, 0 oszlop) állítja. Valójában az adat 
                            //regiszter cím mutatóját állítja 0-ra.
  delayMicroseconds(2000);  //Végrehajtási idő adatlap szerint 1.5msec
}

void LiquidCrystal_I2C_HUN::setCursor(byte oszlop, byte sor){  //Cursor pozíció beállítása, oszlop és sor pl. 0,1 0. oszlop az 1. sorban
  int sor_cimeltolas[] = { 0x00, 0x40, 0x14, 0x54 };           //ezek a különböző sorok első karakter címeinek kiszámításához szükséges növekmények
  if (sor > _lcd_sor) {
    sor = _lcd_sor-1;     // Ha nagyobb sor számot ad meg mint a kijelző soroainak száma, akkor az utolsóba írunk
  }
  command(B10000000 | (oszlop + sor_cimeltolas[sor])); //DDRAM címzése
}

//Kikapcsoljuk a kijelző feliratainak megjelenítését (háttérvilágítás ha van, marad)
void LiquidCrystal_I2C_HUN::noDisplay() {
  _displaycontrol&=B11111011;    //a _displaycontrol változó 3. bitje felel a kijelző be és kikapcsolásért, 0-val kikapcsol
  command(_displaycontrol);
}
void LiquidCrystal_I2C_HUN::display() {
  _displaycontrol|=B00000100;    //a _displaycontrol változó 3. bitje felel a kijelző be és kikapcsolásért, 1-el bekapcsol
  command(_displaycontrol);
}
// Kijelző teljes ki é bekapcsolása (háttérvilágítás is)
void LiquidCrystal_I2C_HUN::off() {                 //Kijelző megjelenítés kikapcsolása, háttérvilágítás lekapcsolása, DDRAM tartalma marad
  _backlightval=B00000000;       //Háttérvilágítást vezérlő változóban a 4. bit felülírása 0-val
  _displaycontrol&=B11111011;    //a _displaycontrol változó 3. bitje felel a kijelző be és kikapcsolásért, 0-val kikapcsol
  command(_displaycontrol);      //Display control parancs kiküldése, ezzel együtt háttérvilágítás kapcsolása is megtörténik
}
void LiquidCrystal_I2C_HUN::on() {                 //Kijelző megjelenítés bekapcsolása, háttérvilágítás lekapcsolása,  DDRAM tartalma megmaradt, utolsó kiírt tartalom megjelenik
  _backlightval=B00001000;      //Háttérvilágítást vezérlő változóban a 4. bit felülírása 1-el
  _displaycontrol|=B00000100;   //a _displaycontrol változó 3. bitje felel a kijelző be és kikapcsolásért, 1-el bekapcsol
  command(_displaycontrol);     //Display control parancs kiküldése, ezzel együtt háttérvilágítás kapcsolása is megtörténik
}

// cursor (aláhúzás) ki és bekapcsolása
void LiquidCrystal_I2C_HUN::noCursor() {
  _displaycontrol &= B11111101;    //A _displaycontrol változó 2. bitje felel a cursor ki és bekapcsolásáért. Cursor ki 0-val!
  command(_displaycontrol);
}
void LiquidCrystal_I2C_HUN::cursor() {
  _displaycontrol |= B00000010;   //A _displaycontrol változó 2. bitje felel a cursor ki és bekapcsolásáért. Cursor be 1-el!
  command(_displaycontrol);
}

// Cursor villogás ki és bekapcsolása
void LiquidCrystal_I2C_HUN::noBlink() {
  _displaycontrol &= B11111110;   //A _displaycontrol változó jobbról 1. bitje felel a cursor villogás ki és bekapcsolásáért. Villogás ki 0-val!
  command(_displaycontrol);
}
void LiquidCrystal_I2C_HUN::blink() {
  _displaycontrol |= B00000001;   //A _displaycontrol változó jobbról 1. bitje felel a cursor villogás ki és bekapcsolásáért. Villogás be 1-el!
  command(_displaycontrol);
}

//Képernyő görgetése és cursor mozgatása a RAM tartalmának megváltoztatása nélkül. Minden sor egyszerre mozog, a 
//második sorban kilépő karakterek (jobbra tolás) belépnek az első sor elején. Balra toláskor az első sor elején kilépő
//karakterek belépnek az utolsó sor végén.
void LiquidCrystal_I2C_HUN::scrollDisplayLeft(void) {
  command(B00011000);
}
void LiquidCrystal_I2C_HUN::scrollDisplayRight(void) {
  command(B00011100);
}

// A szöveg kiírása balról jobbra halad
void LiquidCrystal_I2C_HUN::leftToRight(void) {
  _displaymode |= B00000010;             //a cursort jobbra lépteti egy karakter kiírását követően (DDRAM címet növeli)
  command(B00000100 | _displaymode);
}

// A szöveg kiírása jobbról balra halad
void LiquidCrystal_I2C_HUN::rightToLeft(void) {
  _displaymode &= B11111101;             //a cursort balra lépteti egy karakter kiírását követően (DDRAM címet csökkenti)
  command(B00000100 | _displaymode);
}

// A szöveget a cursor-tól jobbra tolja el (görgeti). A cursor aktuális pozíciója nem változik
void LiquidCrystal_I2C_HUN::autoscroll(void) {
  _displaymode |= B00000001;             //A kijelző tartalmát tolja el (shift) a kurzortól kezdve 
  command(B00000100 | _displaymode);
}

// A szöveget a cursor-tól balra tolja el (görgeti). A cursor aktuális pozíciója nem változik
void LiquidCrystal_I2C_HUN::noAutoscroll(void) {
  _displaymode &= B11111110;             //A cursort mozgatja, és nem a kijelző tartalmat (no shift)
  command(B00000100 | _displaymode);
}

//Feltölti egy karakter kép feltöltését a CGRAM-ba
//Összesen 8 x 5 byte áll rendelkezésre.
void LiquidCrystal_I2C_HUN::createChar(byte tarolo, byte charmap[]) {  //tarolo a ramterület kezdőcíme, charmap a karakterkép
  tarolo &= B00000111;                  //Csak három bit-et vesszük figyelembe (8 tároló)
  command(B01000000 | (tarolo << 3));   //jobbról a 7. bit jelöli ki a parancsot, 6-4. bit a cím
                                        //A CGRAM címének beállítása után az adat küldések (RS=1) a CGRAM-ba íródnak
  for (int i=0; i<8; i++) {             //Adatok írása, amik a megcímzett CGRAM-ba kerülnek (minden írás növeli a címet)
    write(charmap[i]);
  }
}


//Háttérvilágítás ki és bekapcsolása. A háttérvilágítás led-je az I2C illesztő IC 
//P3 portjára van kötve, és nem az LCD vezérlő kapcsolgatja. Ezért ebben az esetben 
//csak a _backlightval változó jobbról 4. bitjét kell beállítani, és kiküldeni
//bárilyen adatot az lcd_i2cWrite() függvénnyel, ami egy vagy kapcsolattal
//a kiküldött adat jobbról 4. bitjét beállítja az I2C IC portján.
void LiquidCrystal_I2C_HUN::noBacklight(void) {   //Háttérvilágítás kikapcsolása
  _backlightval=B00000000;                        //bit beállítása a változóban
  i2cWrite(0);                                    //egy 0-ás adat küldése az I2C-ra, hogy frissüljön 
                                                  //a bit (nincs B00000000 parancs, amit a chip értelmezne)
}
void LiquidCrystal_I2C_HUN::backlight(void) {     //Háttérvilágítás bekapcsolása
  _backlightval=B00001000;                        //bit beállítása a változóban
  i2cWrite(0);                                    //egy 0-ás adat küldése az I2C-ra, hogy frissüljön 
                                                  //a bit (nincs B00000000 parancs, amit a chip értelmezne)
}
void LiquidCrystal_I2C_HUN::setBacklight(byte ertek){   //más könyvtárakkal készült programok könnyebb adaptáláshoz
  if (ertek) {
    backlight();    // bekapcsolja a háttérvilágítást
  } else {
    noBacklight();  // kikapcsolja a háttérvilágítást
  }
}
bool LiquidCrystal_I2C_HUN::getBacklight() {      //_backlightval változó lekérdezése
  return _backlightval == B00001000;
}


//Adat vagy parancs küldés
void LiquidCrystal_I2C_HUN::command(byte ertek) {
  send(ertek, B00000000);         //LCD vezérlő RS kivezetése 0, akkor parancsot küldünk
}
void LiquidCrystal_I2C_HUN::write(byte ertek) {
  send(ertek, B00000001);         //LCD vezérlő RS kivezetése 1, akkor adatot küldünk a DDRAM-ba vagy a CGRAM-ba
  return 1;
}


//Alacsonyszintű írási parancsok. 
//--------------------------------------------------------------------------
// Az adat vagy parancs kiírása az LCD meghajtóra 4 bites módban.
//A kiküldött parancsot alsó és felső 4 bitre kell szétválasztani, 
//és egymás után kiküldeni az LCD meghajtónak (a kiküldés több lépésből 
//áll, mert az En kivezetést közben 1-ről 0-ra kell ejteni, majd vissza 1-re
void LiquidCrystal_I2C_HUN::send(byte value, byte mode) {
  byte highnib=value&0xf0;      //felső négy bit 
  byte lownib=(value<<4)&0xf0;  //alsó négy bit
                                //az I"C meghajtó csip P7-P4 portjára kell ezeket írni
  wr4bit((highnib)|mode);       //felső négy bit küldése
  wr4bit((lownib)|mode);        //alsó négy bit küldése
}

void LiquidCrystal_I2C_HUN::wr4bit(byte ertek) {
  i2cWrite(ertek | B00000100);               //adat beállítása az I2C meghajtó P7-P4 kivezetés, valamint az  
                                             //En kivezetésen magas szint. Ekkor még nem történik semmi.
  delayMicroseconds(1);                      // Miután magasra állítottuk, adatlap szerint várni kell 450nsec-et											 
  //LCD meghajtó En kivezetésén lefutó él előállítása 
  i2cWrite(ertek & B11111011);   // En (P2) alacsony, ekkor olvassa be az adatot az LCD meghajtó
  delayMicroseconds(50);         // Adat beolvasás LCD meghajtóba adatlap szerint 37us, ezt megvárjuk
}                                  

void LiquidCrystal_I2C_HUN::i2cWrite(byte adat) {   //egy byte kiírása az I2C illesztőre. A kiírt adat a P0-P7 kivezetéseken jelenik meg
                                                    //Az I2C illesztő chip P2 kivezetésére van kötve a Led világítás, ezért annak értékét 
                                                    //minden adat átadásakor kezelni kell (vagy kapcsolattal).
  Wire.beginTransmission(_lcd_cim);
  Wire.write((int)(adat) | _backlightval);
  Wire.endTransmission();
}


//különbözű numerikus változó típusok kiírásának függvényei
//------------------------------------------------------------------------------------------------
void LiquidCrystal_I2C_HUN::print(String string) {  //az LCD kijelzőre küld egy karakter tömböt
  byte i=0;
  byte j=0;
  while ((byte)string[i]>8 and j<_lcd_oszlop) {   //A bináris formátum képzése után a 8-as indexű tömbelembe 8-at tesz (lecseréli az általam berakott 0-át)
                                                  //E miatt nem a 0-át figyelem, nem csak 0 lehet a záró karakter.
    j++;
    if ((byte)string[i]==195 or (byte)string[i]==197) {   //Az UTF 8 kódkészletben a magyar ékezetes betűk 195 vagy 197-el kezdődnek, és ezt követi egy 
                                                          //byte-on az ékezetes karakterkép következő byte-ja. Tehát az első byte-ot ki kell hagyni, és csak
                                                          //a következővel kell foglalkozni. A magyar ékezeteket a CGRAM-ból jelenítjük meg, ezért a megfelelő
                                                          //0-7 tartományban lévő kódra kell cserélni a második byte-on érkező kódot. 
      i++;                                                //A tömbnek ezt az elemét nem írjuk a kijelzőre, a következő tömbelemet vizsgáljuk és kiírjuk a megfelelő karaktert
      switch ((byte)string[i]) {
        case 129: write(0);break;    //Á  
        case 161: write(0);break;    //á
        case 137: write(1);break;    //É
        case 169: write(1);break;    //é
        case 141: write(2);break;    //Í
        case 173: write(2);break;    //í
        case 147: write(3);break;    //Ó
        case 179: write(3);break;    //ó
        case 150: write(239);break;  //Ö
        case 182: write(239);break;  //ö
        case 144: write(4);break;    //Ő
        case 145: write(4);break;    //ő
        case 154: write(5);break;    //Ú
        case 186: write(5);break;    //ú
        case 156: write(245);break;  //Ü
        case 188: write(245);break;  //ü
        case 176: write(6);break;    //Ű
        case 177: write(6);break;    //ű
      }
    }
    else {write((byte)string[i]);}
    i++;
  }
}

void LiquidCrystal_I2C_HUN::print(byte ertek) {   //egy integer értéket alakít át karaktertömbbe, és ezt küldi az lcd kijelzőre 
  char string[_lcd_oszlop];
  sprintf(string, "%d", ertek);
  print(string);
}

void LiquidCrystal_I2C_HUN::print(byte ertek,byte formatum) {   //egy integer értéket alakít át karaktertömbbe, és ezt küldi az lcd kijelzőre 
  char string[_lcd_oszlop];
  switch (formatum) {
    case 16:
      sprintf(string, "%X", ertek);break;
    case 10:
      sprintf(string, "%d", ertek);break;
    case 2:
      byte j=128;
      for(byte i=0;i<7;i++) {           
        if(ertek>=j){ertek-=j;string[i]= '1';}
        else {string[i]='0';}
        j=j/2;
      }
      string[8]=0;break;
  }
  print(string);   //a string átadásakor valamiért dec 8 kerül a 8. tömbelembe. Nem értem miért?
}

void LiquidCrystal_I2C_HUN::print(int ertek) {            //egy integer értéket alakít át karaktertömbbe, és ezt küldi az lcd kijelzőre 
  char string[_lcd_oszlop];
  sprintf(string, "%d", ertek);
  print(string);
}

void LiquidCrystal_I2C_HUN::print(unsigned int ertek) {   //egy unsigned integer értéket alakít át karaktertömbbe, és ezt küldi az lcd kijelzőre 
  char string[_lcd_oszlop];
  sprintf(string, "%u", ertek);
  print(string);
}

void LiquidCrystal_I2C_HUN::print(long ertek) {            //egy long értéket alakít át karaktertömbbe, és ezt küldi az lcd kijelzőre 
  char string[_lcd_oszlop];
  Serial.println("long");
  sprintf(string, "%ld", ertek);
  print(string);
}

void LiquidCrystal_I2C_HUN::print(unsigned long ertek) {   //egy unsigned long értéket alakít át karaktertömbbe, és ezt küldi az lcd kijelzőre 
  char string[_lcd_oszlop];
  sprintf(string, "%lu", ertek);
  print(string);
}

void LiquidCrystal_I2C_HUN::print(float ertek, byte tizedes) {   //egy float értéket alakít át karaktertömbbe, és ezt küldi az lcd kijelzőre 
                                                                 //második paraméterként megadható a tizedes jegyek száma (kötelező megadni)!
  char string[_lcd_oszlop];
  dtostrf(ertek, 1, tizedes, string);  
  print(string);
}
